import unittest
import numpy as np


class TestCreatePfp(unittest.TestCase):
    def test_create_pfp_default(self):
        from polymerfingerprint import create_pfp

        repeating_units = {"C=C": 0.5, "C#C": 0.5}
        mol_weight = 150.0

        pfp = create_pfp(repeating_units, mol_weight)
        # Default settings should produce a Polyfingerprint with a length of:
        # 2048 + (2 * 2048) * 2 = 10240
        self.assertIsInstance(pfp, np.ndarray)
        self.assertEqual(len(pfp), 10240)

    def test_create_pfp_with_enhanced_functions(self):
        from polymerfingerprint.fingerprints import create_RDKFingerprint
        from polymerfingerprint import create_pfp

        repeating_units = {"C=C": 0.5, "C#C": 0.5}
        mol_weight = 150.0
        enhanced_fp_functions = [
            create_RDKFingerprint,
            create_RDKFingerprint,
        ]  # Using the same function twice for simplicity

        pfp = create_pfp(
            repeating_units, mol_weight, enhanced_fp_functions=enhanced_fp_functions
        )
        # With the above settings, the Polyfingerprint should have a length of:
        # 2048 + (2 * 2048) * 2 = 10240
        self.assertIsInstance(pfp, np.ndarray)
        self.assertEqual(len(pfp), 10240)


if __name__ == "__main__":
    unittest.main()



import unittest
from rdkit import Chem
from rdkit.Chem import Descriptors
import numpy as np


def molar_wt_from_smiles(smiles_1):
    pass


class TestmolarWtfromSmiles(unittest.TestCase):
    def test_molar_wt_from_smiles_1(self):
        from polymerfingerprint.utils import molar_wt_from_smiles

        smiles_1 = "C"  # Methane
        smiles_2 = "CC"  # Water
        mol = Chem.MolFromSmiles(smiles_1)
        expected_mol_weight = Descriptors.MolWt(mol)
        calculated_mol_weight = molar_wt_from_smiles(smiles_1)
        self.assertAlmostEqual(expected_mol_weight, calculated_mol_weight, places=5)

        mol = Chem.MolFromSmiles(smiles_2)
        expected_mol_weight = Descriptors.MolWt(mol)
        calculated_mol_weight = molar_wt_from_smiles(smiles_2)
        self.assertAlmostEqual(expected_mol_weight, calculated_mol_weight, places=5)

    def test_molar_wt_fromSmiles_invalid_smiles(self):
        from polymerfingerprint.utils import molar_wt_from_smiles

        # This is to test if the function behaves well with an invalid smiles string
        invalid_smiles = "invalid_smiles"
        with self.assertRaises(Exception):
            molar_wt_from_smiles(invalid_smiles)


class TestCalcPolymerShares(unittest.TestCase):
    def setUp(self):
        # Sample repeating units and end groups
        self.rep_units = {"C": 0.6, "O": 0.4}
        self.ends = ["F", "Cl"]
        self.total_weight = 100

    def test_polymer_shares_give_100perc(self):
        from polymerfingerprint.utils import calc_polymer_shares

        ru_mole_fractions, ends_mole_fraction = calc_polymer_shares(
            self.rep_units, self.ends, self.total_weight
        )

        self.assertTrue(sum(ru_mole_fractions.values()) + sum(ends_mole_fraction) == 1)

    def test_total_weight_less_than_ends(self):
        from polymerfingerprint.utils import calc_polymer_shares

        with self.assertRaises(ValueError):
            calc_polymer_shares(self.rep_units, ["F", "CCCCC"], 50)

    def test_normalize_rep_units(self):
        from polymerfingerprint.utils import calc_polymer_shares

        rep_units = {"C": 60, "O": 40}
        ru_mole_fractions, _ = calc_polymer_shares(
            rep_units, self.ends, self.total_weight
        )
        self.assertTrue(sum(ru_mole_fractions.values()) <= 1)

    def test_resulting_mole_fractions_sum_to_1(self):
        from polymerfingerprint.utils import calc_polymer_shares

        ru_mole_fractions, ends_mole_fraction = calc_polymer_shares(
            self.rep_units, self.ends, self.total_weight
        )

        self.assertAlmostEqual(
            sum(ru_mole_fractions.values()) + sum(ends_mole_fraction), 1, places=5
        )


class TestRepeatingUnitCombinations(unittest.TestCase):
    def test_combinations(self):
        from polymerfingerprint.utils import repeating_unit_combinations

        repeating_units = ["[CH2]C[CH2]", "[CH2][CH](-C)"]
        start_unit = "[H]"
        end_unit = "[CH3]"
        expected_combinations = [
            "[H][CH2]C[CH2][CH2]C[CH2][CH3]",
            "[H][CH2]C[CH2][CH2][CH](-C)[CH3]",
            "[H][CH2][CH](-C)[CH2]C[CH2][CH3]",
            "[H][CH2][CH](-C)[CH2][CH](-C)[CH3]",
        ]

        result = repeating_unit_combinations(
            repeating_units, start=start_unit, end=end_unit
        )
        self.assertListEqual(result, expected_combinations)

    def test_combinations_length(self):
        from polymerfingerprint.utils import repeating_unit_combinations

        for i in range(1, 5):
            units = [f"U{j + 1}" for j in range(i)]
            combinations = repeating_unit_combinations(units)
            self.assertEqual(len(combinations), i**i)


class TestTestPolymerSmiles(unittest.TestCase):
    def test_valid_polymer_smiles(self):
        from polymerfingerprint.utils import test_polymer_smiles

        # Provided examples
        self.assertFalse(test_polymer_smiles("CCC(=O)OC"))  # has no open ends
        self.assertFalse(
            test_polymer_smiles("[CH2][CH]C(=O)OC")
        )  # has both open ends, but not at the terminal atoms
        self.assertTrue(
            test_polymer_smiles("[CH2][CH](C(=O)(OC))")
        )  # has both open ends at the terminal atoms

        # Additional test cases
        self.assertFalse(test_polymer_smiles(""))  # empty string
        self.assertTrue(
            test_polymer_smiles("[CH2][CH](C(=O)(OC(C)(C)OCC))")
        )  # a more complex structure with open ends

        # Test cases that can potentially raise exceptions
        self.assertFalse(
            test_polymer_smiles("[CH2][CH]C(=O)(OC")
        )  # unbalanced parentheses
        self.assertFalse(
            test_polymer_smiles("invalidSmiles")
        )  # not a valid SMILES string

    def test_valid_startgroup(self):
        from polymerfingerprint.utils import test_startgroup

        # Provided examples
        self.assertFalse(test_startgroup("CCC"))  # has no open ends
        self.assertFalse(
            test_startgroup("C[CH]C")
        )  # has a radical but not at the terminal atom
        self.assertTrue(
            test_startgroup("C[CH](C)")
        )  # has a radical at the terminal atom

        # Additional test cases
        self.assertFalse(test_startgroup(""))  # empty string
        self.assertFalse(
            test_startgroup("[CH2]CC")
        )  # radical at the start of the molecule
        self.assertTrue(
            test_startgroup("[CH2][CH2]")
        )  # radicals at the end start is optional
        self.assertTrue(test_startgroup("CC[CH2]"))  # radical at the end

        # Test cases that can potentially raise exceptions
        self.assertFalse(test_startgroup("C[CH2]C("))  # unbalanced parentheses
        self.assertFalse(test_startgroup("invalidSmiles"))  # not a valid SMILES string

    def test_valid_endgroup(self):
        from polymerfingerprint.utils import test_endgroup

        # Provided examples
        self.assertFalse(test_endgroup("CCC"))  # has no open ends
        self.assertFalse(
            test_endgroup("C[CH]C")
        )  # has a radical but not at the first atom
        self.assertTrue(test_endgroup("[CH]C(C)"))  # has a radical at the terminal atom

        # Additional test cases
        self.assertFalse(test_endgroup(""))  # empty string
        self.assertFalse(test_endgroup("CC[CH2]"))  # radical at the end of the molecule
        self.assertTrue(
            test_endgroup("[CH2][CH2]")
        )  # radicals at the end start is optional
        self.assertTrue(test_endgroup("[CH2]CC"))  # radical at the start

        # Test cases that can potentially raise exceptions
        self.assertFalse(test_endgroup("[CH2]CC("))  # unbalanced parentheses
        self.assertFalse(test_endgroup("invalidSmiles"))  # not a valid SMILES string


class TestCategoricalFunction(unittest.TestCase):
    def test_with_categorical_data(self):
        from polymerfingerprint.utils import test_categorical

        self.assertTrue(test_categorical(["apple", "banana", "cherry"]))
        self.assertTrue(
            test_categorical(np.array(["apple", "banana", "cherry"], dtype=object))
        )
        self.assertTrue(test_categorical(np.array(["1", "2", "3"], dtype=str)))

    def test_with_numerical_data(self):
        from polymerfingerprint.utils import test_categorical

        self.assertFalse(test_categorical([1, 2, 3]))
        self.assertFalse(test_categorical(np.array([1.5, 2.5, 3.5])))
        self.assertFalse(test_categorical(np.array([1, 2, 3], dtype=np.int64)))
        self.assertFalse(test_categorical(np.array([1, 2, 3], dtype=np.uint8)))
        self.assertFalse(test_categorical(np.array([True, False, True], dtype=bool)))

    def test_with_empty_data(self):
        from polymerfingerprint.utils import test_categorical

        with self.assertRaises(ValueError):
            test_categorical([])

    def test_with_higher_dimensional_data(self):
        from polymerfingerprint.utils import test_categorical

        with self.assertRaises(ValueError):
            test_categorical([[1, 2, 3], [4, 5, 6]])

    def test_with_pandas_series(self):
        from polymerfingerprint.utils import test_categorical
        import pandas as pd

        self.assertTrue(test_categorical(pd.Series(["apple", "banana", "cherry"])))
        self.assertFalse(test_categorical(pd.Series([1, 2, 3])))


if __name__ == "__main__":
    unittest.main()



from typing import List, Optional
import pandas as pd
import numpy as np

from rdkit import Chem

import polymerfingerprint as pfp

REPEATING_UNIT_COLUMN_PREFIX = "SMILES_repeating_unit"
MOLPERCENT_RU_COLUMN_PREFIX = "molpercent_repeating_unit"

ADDITIVE_COLUMN_PREFIX = "additive"
ADDITIVE_WP_COLUMN_POSTFIX = "_weight_percent"
ADDITIVE_CONC_COLUMN_POSTFIX = "_concentration_molar"

SMILES_START_GROUP_COL = "SMILES_start_group"
SMILES_END_GROUP_COL = "SMILES_end_group"
MN_COL = "Mn"


class InvalidSmilesError(Exception):
    pass


def check_additive(additive, weight_percent=None, concentration=None, density=1):
    if pd.isnull(additive):
        return {}

    if pd.isnull(weight_percent):
        weight_percent = None
    if pd.isnull(concentration):
        concentration = None

    mol = Chem.MolFromSmiles(additive)
    if mol is None:
        raise InvalidSmilesError(f"additive {additive} is not a valid SMILES")

    if weight_percent is not None:
        conc = weight_percent / Chem.Descriptors.MolWt(mol) * 1000 * density
    if concentration is not None:
        if weight_percent is not None:
            if not np.isclose(concentration, conc, rtol=1e-2, atol=1e-3):
                raise ValueError(
                    f"concentration {concentration} and weight_percent {weight_percent} (conc={conc}) do not match"
                )

        conc = concentration

    # seperate cations and anions

    frags = list(Chem.GetMolFrags(mol, asMols=True))
    smiles = [Chem.MolToSmiles(frag) for frag in frags]
    unique_smiles = {}
    for s in smiles:
        if s not in unique_smiles:
            unique_smiles[s] = 1
        else:
            unique_smiles[s] += 1

    return {k: v * conc for k, v in unique_smiles.items()}


def expand_data(
    sampledf,
    ignored_columns: Optional[List[str]] = None,
    density=1,
    repeating_unit_column_prefix=REPEATING_UNIT_COLUMN_PREFIX,
    molpercent_ru_column_prefix=MOLPERCENT_RU_COLUMN_PREFIX,
    additive_column_prefix=ADDITIVE_COLUMN_PREFIX,
    additive_wp_column_postfix=ADDITIVE_WP_COLUMN_POSTFIX,
    additive_conc_column_postfix=ADDITIVE_CONC_COLUMN_POSTFIX,
    smiles_start_group_col=SMILES_START_GROUP_COL,
    smiles_end_group_col=SMILES_END_GROUP_COL,
    mn_col=MN_COL,
):
    if mn_col not in sampledf.columns:
        raise ValueError(f"Column '{mn_col}' not found in excel file.")
    if ignored_columns is None:
        ignored_columns = []

    # drop ingored columns
    for c in ignored_columns:
        if c in sampledf.columns:
            sampledf = sampledf.drop(columns=[c])

    unused_columns = list(sampledf.columns)
    unused_columns.remove(mn_col)

    # remove samples without mn
    index_with_nan_mn = sampledf.index[sampledf[mn_col].isna()].tolist()
    sampledf = sampledf.drop(index=index_with_nan_mn)

    # extract repeating units
    repeating_unit_columns = []
    for col in sampledf.columns:
        if not col.startswith(repeating_unit_column_prefix):
            continue
        c = col[len(repeating_unit_column_prefix) :]
        if molpercent_ru_column_prefix + c not in sampledf.columns:
            pfp.PFPLOGGER.warning(
                f"Missing molpercent column for repeating unit {c}."
                f" Skipping this repeating unit and using its columns as datapoint"
            )
            continue
        unused_columns.remove(repeating_unit_column_prefix + c)
        unused_columns.remove(molpercent_ru_column_prefix + c)
        repeating_unit_columns.append(c)

    unique_rus = set()
    for c in repeating_unit_columns:
        notnanru = sampledf[repeating_unit_column_prefix + c][
            ~sampledf[repeating_unit_column_prefix + c].isna()
        ]
        if len(notnanru) > 0:
            unique_rus.add(notnanru.unique()[0])
    # check if all repeating units valid
    invalid_rus = [s for s in unique_rus if not pfp.test_polymer_smiles(s)]
    invlaid_rus_indices = []
    if invalid_rus:
        for c in repeating_unit_columns:
            invlaid_rus_indices.extend(
                sampledf.index[
                    sampledf[repeating_unit_column_prefix + c].isin(invalid_rus)
                ].tolist()
            )
    sampledf = sampledf.drop(index=invlaid_rus_indices)

    if smiles_start_group_col not in sampledf.columns:
        sampledf[smiles_start_group_col] = "[H]"
    if smiles_start_group_col in unused_columns:
        unused_columns.remove(smiles_start_group_col)
    unique_start_groups = set(sampledf[smiles_start_group_col].unique())
    invalid_start_groups = [
        s for s in unique_start_groups if not pfp.test_startgroup(s)
    ]
    invalid_start_groups_indices = []
    if invalid_start_groups:
        invalid_start_groups_indices = sampledf.index[
            sampledf[smiles_start_group_col].isin(invalid_start_groups)
        ].tolist()
        sampledf = sampledf.drop(index=invalid_start_groups_indices)

    if smiles_end_group_col not in sampledf.columns:
        sampledf[smiles_end_group_col] = "[H]"
    if smiles_end_group_col in unused_columns:
        unused_columns.remove(smiles_end_group_col)
    unique_end_groups = set(sampledf[smiles_end_group_col].unique())
    invalid_end_groups = [s for s in unique_end_groups if not pfp.test_endgroup(s)]
    invalid_end_groups_indices = []
    if invalid_end_groups:
        invalid_end_groups_indices = sampledf.index[
            sampledf[smiles_end_group_col].isin(invalid_end_groups)
        ].tolist()
        sampledf = sampledf.drop(index=invalid_end_groups_indices)

    additive_columns = []
    for col in sampledf.columns:
        if not col.startswith(additive_column_prefix):
            # skip non-additive columns
            continue
        if col.endswith(additive_conc_column_postfix) or col.endswith(
            additive_wp_column_postfix
        ):
            # skip concentration and weight percent columns
            continue
        c = col[len(additive_column_prefix) :]
        if not c:
            # skip unnamed additive columns
            continue
        molar_c_col = f"{additive_column_prefix}{c}{additive_conc_column_postfix}"
        wp_col = f"{additive_column_prefix}{c}{additive_wp_column_postfix}"

        if wp_col in unused_columns:
            unused_columns.remove(wp_col)
        if wp_col not in sampledf.columns:
            sampledf[wp_col] = np.nan

        if molar_c_col in unused_columns:
            unused_columns.remove(molar_c_col)
        if molar_c_col not in sampledf.columns:
            sampledf[molar_c_col] = np.nan
        additive_columns.append(c)
        if col in unused_columns:
            unused_columns.remove(col)

    # check additive columns
    removed_because_no_additive_amount = []
    for ac in additive_columns:
        additive_not_nan = sampledf[~sampledf[additive_column_prefix + ac].isna()]
        additive_molpercent = additive_not_nan[
            additive_column_prefix + ac + additive_wp_column_postfix
        ]
        additive_weightpercent = additive_not_nan[
            additive_column_prefix + ac + additive_conc_column_postfix
        ]
        # inces of rows where no additive amount is given in additive_molpercent or additive_weightpercent
        removed_because_no_additive_amount = additive_not_nan.index[
            additive_molpercent.isna() & additive_weightpercent.isna()
        ].tolist()
    sampledf = sampledf.drop(index=set(removed_because_no_additive_amount))

    # split additives into
    additive_splits = {i: {} for i in sampledf.index}
    removed_because_invalid_additive_smiles = []
    removed_because_invalid_additive_amount = []
    for index, row in sampledf.iterrows():
        for ac in additive_columns:
            additive_smiles = row[additive_column_prefix + ac]
            additive_weightpercent = row[
                additive_column_prefix + ac + additive_wp_column_postfix
            ]
            additive_conc = row[
                additive_column_prefix + ac + additive_conc_column_postfix
            ]
            try:
                additive_split = check_additive(
                    additive_smiles,
                    weight_percent=additive_weightpercent,
                    concentration=additive_conc,
                    density=density,
                )
                for smiles, amount in additive_split.items():
                    if smiles in additive_splits[index]:
                        additive_splits[index][smiles] += amount
                    else:
                        additive_splits[index][smiles] = amount
            except InvalidSmilesError:
                pfp.PFPLOGGER.exception(
                    f"Invalid SMILES {additive_smiles} in row {index}. Skipping additive."
                )
                removed_because_invalid_additive_smiles.append(index)
            except ValueError as e:
                pfp.PFPLOGGER.warning(str(e) + f" in row {index}. Skipping additive.")
                removed_because_invalid_additive_amount.append(index)

    sampledf = sampledf.drop(
        index=set(
            removed_because_invalid_additive_smiles
            + removed_because_invalid_additive_amount
        )
    )

    for ac in additive_columns:
        sampledf = sampledf.drop(columns=[additive_column_prefix + ac])
        sampledf = sampledf.drop(
            columns=[additive_column_prefix + ac + additive_conc_column_postfix]
        )
        sampledf = sampledf.drop(
            columns=[additive_column_prefix + ac + additive_wp_column_postfix]
        )

    unique_additives = set()
    for split in additive_splits.values():
        unique_additives.update(split.keys())
    unique_additives = list(unique_additives)

    additive_cols = []
    for i, additive in enumerate(unique_additives):
        sampledf[f"additive_{i}"] = 0.0
        additive_cols.append(f"additive_{i}")

    for index, row in sampledf.iterrows():
        additive_split = additive_splits[index]
        for smiles, amount in additive_split.items():
            sampledf.loc[index, f"additive_{unique_additives.index(smiles)}"] = amount

    numerical_columns = []
    categorical_columns = []
    for col in unused_columns + additive_cols:
        if pfp.utils.test_categorical(sampledf[col]):
            categorical_columns.append(col)
        else:
            numerical_columns.append(col)

    missing_numerical_indices = {}
    for col in numerical_columns:
        nan_indices = sampledf.index[sampledf[col].isna()].tolist()
        sampledf = sampledf.drop(index=nan_indices)
        if len(nan_indices) > 0:
            missing_numerical_indices[col] = nan_indices

    categorical_unique_values = {}
    for catcol in categorical_columns:
        unique_values = list(sampledf[catcol].unique())
        categorical_unique_values[catcol] = unique_values

    info = {
        "repeating_unit_columns": [
            f"{repeating_unit_column_prefix}{c}" for c in repeating_unit_columns
        ],
        "molpercent_repeating_unit_columns": [
            f"{molpercent_ru_column_prefix}{c}" for c in repeating_unit_columns
        ],
        "additive_columns": additive_columns,
        "missing_additive_amount": removed_because_no_additive_amount,
        "invalid_end_groups_indices": invalid_end_groups_indices,
        "invalid_start_groups_indices": invalid_start_groups_indices,
        "invalid_repeating_unit_indices": invlaid_rus_indices,
        "invalid_additive_smiles_indices": removed_because_invalid_additive_smiles,
        "invalid_additive_amount_indices": removed_because_invalid_additive_amount,
        "missing_mn_indices": index_with_nan_mn,
        "additives": unique_additives,
        "numerical_columns": numerical_columns,
        "categorical_columns": categorical_columns,
        "categorical_unique_values": categorical_unique_values,
        "missing_numerical_indices": missing_numerical_indices,
    }

    return sampledf, info


def excel_to_data(path: str, ignored_columns: Optional[List[str]] = None):
    sampledf = pd.read_excel(path, sheet_name="samples")

    datadf = pd.read_excel(path, sheet_name="data")
    datadf = datadf.set_index("property")
    density = datadf.loc["density"]["value"]
    return expand_data(sampledf, ignored_columns=ignored_columns, density=density)



from typing import List, Tuple
import numpy as np
from rdkit import Chem
from .utils import polymol_fom_smiles
from .logger import PFPLOGGER


def create_AtomicPairFingerprint(
    smiles_s: List[str], fp_size: int = 2048, complement: bool = False
) -> List[np.ndarray]:
    """
    Creates a list of AtomicPairFingerprints from a list of SMILES strings.

    Args:
        smiles_s (List[str]): List of SMILES strings.
        fp_size (int, optional): Size of the fingerprint. Defaults to 2048.
        complement (bool, optional): If True, the SMILES strings are patched
            with [H][H] at the beginning and end, should be true if working
            with repeating unit smiles to hide the radicals. Defaults to False.
    Returns:
        List[np.array]: List of AtomicPairFingerprints.
    """
    if complement:
        smiles_list = [("[H]{}[H]".format(smiles)) for smiles in smiles_s]
    else:
        smiles_list = smiles_s

    fingerprint_s = [
        np.array(
            list(
                Chem.rdFingerprintGenerator.GetAtomPairGenerator(
                    fpSize=fp_size
                ).GetFingerprint(polymol_fom_smiles(smiles))
            )
        )
        for smiles in smiles_list
    ]
    return fingerprint_s


def create_RDKFingerprint(
    smiles_s: List[str], fp_size: int = 2048, complement: bool = False
) -> List[np.ndarray]:
    """
    Creates a list of RDKFingerprints from a list of SMILES strings.

    Args:
        smiles_s (List[str]): List of SMILES strings.
        fp_size (int, optional): Size of the fingerprint. Defaults to 2048.
        complement (bool, optional): If True, the SMILES strings are patched
            with [H][H] at the beginning and end, should be true if working
            with repeating unit smiles to hide the radicals. Defaults to False.
    Returns:
        List[np.array]: List of RDKFingerprints.
    """
    if complement:
        smiles_list = [("[H]{}[H]".format(smiles)) for smiles in smiles_s]
    else:
        smiles_list = smiles_s

    fingerprint_s = [
        np.array(Chem.RDKFingerprint(polymol_fom_smiles(smiles), fpSize=fp_size))
        for smiles in smiles_list
    ]
    return fingerprint_s


def merge_bit_fingerprints(
    fingerprints: List[np.ndarray[[-1], bool]],
) -> np.ndarray[[-1], bool]:
    """merges an arbitrary number of bit fingerprints into one by using the or operator

    Args:
        fingerprints (List[np.ndarray[[-1], bool]]): arbitrary number of bit fingerprints with the same length L

    Returns:
        np.ndarray[[-1], bool]: merged fingerprint with length L
    """

    # flatten
    fingerprints = [fp.flatten() for fp in fingerprints]
    # make sure all fingerprints have the same length
    if len(set([len(fp) for fp in fingerprints])) > 1:
        raise ValueError("All fingerprints must have the same length.")

    # make sure all fingerprints are bit fingerprints
    fingerprints = [fp.astype(bool) for fp in fingerprints]

    # merge fingerprints
    merged_fp = np.stack(fingerprints)
    merged_fp = np.any(merged_fp, axis=0)
    return merged_fp


def weight_sum_fingerprints(
    fingerprints: List[np.ndarray[[-1], float]], weights: List[float]
) -> np.ndarray[[-1], float]:
    """sums up a list of fingerprints with weights and returns the weighted sum fingerprint

    Args:
        fingerprints (List[np.ndarray[[-1], float]]): list of fingerprints
        weights (List[float]): list of weights

    Returns:
        np.ndarray[[-1], float]: weighted sum fingerprint

    Raises:
        ValueError: if the number of weights is not the same as the number of fingerprints
        ValueError: if the fingerprints do not have the same length

    Example:
        >>> weight_sum_fingerprints([np.array([1,2,3]), np.array([4,5,6])], [0.5, 0.5])
        np.array([2.5, 3.5, 4.5])
    """

    # flatten
    fingerprints = [fp.flatten() for fp in fingerprints]
    # make sure all fingerprints have the same length L
    if len(set([len(fp) for fp in fingerprints])) > 1:
        raise ValueError("All fingerprints must have the same length.")

    # make sure all weights have the same length as the number of fingerprints
    if len(weights) != len(fingerprints):
        raise ValueError(
            "Number of weights must be the same as the number of fingerprints."
        )

    fingerprints = [fp * weights[i] for i, fp in enumerate(fingerprints)]

    return np.sum(fingerprints, axis=0)


def reduce_fp_set(
    fingerprints: List[np.ndarray[[-1], float]], check_correlation: bool = True
) -> Tuple[
    List[np.ndarray[[-1], float]], np.ndarray[[-1], bool], np.ndarray[[-1], float]
]:
    """
    Reduces a set of fingerprints by removing positions that have identical values across all provided fingerprints.

    Given multiple fingerprints, this function identifies and discards the positions (features)
    that are consistent among all the fingerprints.Secondly it keeps only the first of those position which are
    correlating completely positively or negatively, as they deemed as less informative for certain analyses,
    since they don't contribute to the distinction between fingerprints.

    Args:
        fingerprints (List[np.ndarray]): A list of 1D numpy arrays representing the fingerprints.
        check_correlation (bool, optional): If True, the function will also remove positions that correlate completely
        positive or negatively. Defaults to True.

    Returns:
        Tuple[List[np.ndarray], np.ndarray, np.ndarray]:
            - List[np.ndarray]: A list of reduced 1D numpy arrays where identical positions across all fingerprints have
                been removed.
            - np.ndarray: A boolean mask indicating the positions that were kept (False) or removed (True).
            - np.ndarray: The first fingerprint from the input, prior to any reductions.

    Note:
        This function assumes that all input fingerprints are of the same length. It also logs the percentage
        reduction in fingerprint size, which might be useful for understanding the impact of the reduction on the data.

    Examples: # position 3 is correlating with position 1 (1 is kept) and position 4 is never changing
        >>> fp1 = np.array([0.2, 0.5, 0.2, 0.0])
        >>> fp2 = np.array([0.4, 0.6, 0.4, 0.0])
        >>> fp3 = np.array([0.2, 0.7, 0.2, 0.0])
        >>> reduced_fps, mask, reference_fp = reduce_fp_set(fp1, fp2, fp3)
        >>> print(reduced_fps)  # Lists of reduced fingerprints
        np.array([[0.2, 0.5], [0.4, 0.6], [0.2, 0.7]])
        >>> print(mask)         # Mask used for reduction
        np.array([False, False, True, True])
        >>> print(reference_fp) # Reference fingerprint
        np.array([0.2, 0.5, 0.2, 0.0])
    """

    # Stack the fingerprints to identify common positions
    stacked_fps = np.stack(fingerprints)

    # Identify positions that are the same across all fingerprints
    same_positions = np.all(stacked_fps == stacked_fps[0, :], axis=0)
    # This is a mask for all positions that should be excluded

    # Reduce the fingerprints using the mask
    reduced_fps = stacked_fps[:, ~same_positions]

    if check_correlation:
        # calculate correlation matrix from the transposed fp data
        correlation_matrix = np.corrcoef(reduced_fps.T)

        # Determine columns to keep
        threshold = 0.999999  # Threshold is basically 1 but not exactly due to floating point errors
        n_cols = correlation_matrix.shape[0]
        keep = []

        # Keep track of columns already processed
        processed = set()

        for i in range(n_cols):
            # Skip columns that have already been processed
            if i in processed:
                continue
            # Keep the current column
            keep.append(i)
            for j in range(i + 1, n_cols):
                # If the correlation between columns i and j is above the threshold, mark column j as processed
                if abs(correlation_matrix[i, j]) > threshold:
                    processed.add(j)

        # reduce with the correlation
        reduced_fps = reduced_fps[:, keep]  # keep columns collected in the loop

        # combine the mask with the correlation mask
        mask = same_positions | np.array(
            [i not in keep for i in range(len(same_positions))]
        )

    PFPLOGGER.info(
        "reduced size by {0:.0f}%".format(
            (1 - (len(reduced_fps[0]) / len(fingerprints[0]))) * 100
        )
    )

    # Return the reduced fingerprints, the mask, and a reference fingerprint
    return reduced_fps, mask, fingerprints[0].copy()


def apply_reduction_fp_set(
    fingerprints: List[np.ndarray[[-1], float]],
    mask: np.ndarray[[-1], bool],
    reference_fp: np.ndarray[[-1], float],
) -> List[np.ndarray[[-1], float]]:
    """
    Given multiple fingerprints, this function discards the same positions (features)
    like the reduce_fp_set function did with a prior set, thus rendering the fingerprints comparable.

    Args:
        fingerprints (List[np.ndarray]): A list of 1D numpy arrays representing the fingerprints.
        mask (np.ndarray): A boolean mask indicating the positions that were kept (False) or removed (True).
        reference_fp (np.ndarray): The first fingerprint from the input, prior to any reductions
            (needed for the information loss calculation).

    Returns:
        Tuple[List[np.ndarray], np.ndarray, np.ndarray]:
            - List[np.ndarray]: A list of 1D numpy arrays reduced by the mask.

    Note:
        This function assumes that all input fingerprints and the maske (prior set) are of the same length.
        It also logs the loss which depicts the percentage of information lost by the reduction, respectively positions
         that were new but removed for comparability.

    Examples:
        >>> fp4 = np.array([0.2, 0.6, 0.2])
        >>> fp5 = np.array([0.2, 0.7, 0.1])
        >>> mask = np.array([True, False, True])
        >>> reference_fp = np.array([0.2, 0.5, 0.1])
        >>> reduced_fps= apply_reduction_fp_set([fp4, fp5], mask, reference_fp)
        >>> print(reduced_fps)  # Lists of reduced fingerprints
        np.array([[[0.6], [0.7]])
    """
    is_out_count = mask.sum()
    if is_out_count == 0:
        return fingerprints

    stacked_fps = np.stack(fingerprints)
    should_out = (stacked_fps == reference_fp) & mask
    should_out_count = should_out.sum(1)
    min_should_out = should_out_count.min()
    mean_should_out = should_out_count.mean()

    PFPLOGGER.info(
        "mean reduction loss is %.0f%% with the highest loss per fingerprint beeing %.0f%%",
        (1 - mean_should_out / is_out_count) * 100,
        (1 - min_should_out / is_out_count) * 100,
    )

    return [new_fp[~mask] for new_fp in fingerprints]



from typing import List, TypedDict, Optional, Union, Callable, Dict
import numpy as np


class PfpData(TypedDict):
    """TypedDict containing the data for a single Polyfingerprint.

    Args:
        repeating_units (Dict[str: float]): Dictionary containing the
            SMILES representation of each repeating unit and the
            corresponding relative amount.
    """

    repeating_units: Dict[str, float]
    startgroup: Optional[str]
    endgroup: Optional[str]
    y: Optional[Union[float, int]]
    mw: float
    pfp: Optional[np.ndarray]


FingerprintFunction = Callable[[List[str], int, bool], List[np.ndarray]]
"""
Type hint for the fingerprint creation functions.

The function expects the following parameters:
    - A list of strings (SMILES strings).
    - An integer (the fingerprint size), which defaults to 2048.
    - A boolean (whether to complement the SMILES strings or not), which defaults to False.

The function returns:
    - A list of numpy arrays (the fingerprints).
"""



import logging

PFPLOGGER = logging.getLogger("polyfingerprints")
PFPLOGGER.setLevel(logging.INFO)

handler = logging.StreamHandler()
handler.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
PFPLOGGER.addHandler(handler)



from typing import Tuple, List, Optional, Dict, Iterable
from rdkit.Chem import Descriptors, Mol, MolFromSmiles
import numpy as np
from rdkit import RDLogger


def molar_wt_from_smiles(smiles: str) -> float:
    return Descriptors.MolWt(polymol_fom_smiles(smiles))


def calc_polymer_shares(
    rep_units: dict[str, float],
    start_end_smiles: List[str],
    total_weight: float,
) -> Tuple[Dict[str, float], List[float]]:
    """calculates the mole fraction of the polymer repeating units and the start end groups

    Args:
        rep_units (Dict[str:float]): Dictionary containing the SMILES representation of each repeating unit and the
            corresponding relative amount.
        ends (List[str]): list of SMILES strings for the start and end groups

    Returns:
        Tuple[Dict[str:float], List[float]]: Tuple containing the mole fraction of the repeating units and
            the start end groups

    """

    total_weight_minus = total_weight - sum(
        molar_wt_from_smiles(x) for x in start_end_smiles
    )
    if total_weight_minus < 0:
        raise ValueError(
            "Total weight of the polymer is smaller than the weight of the start and end groups."
        )

    # normalize rep_units, normally this should be done in a previous step, but just in case
    total_amount = sum([ru for ru in rep_units.values()])
    rep_units = {k: v / total_amount for k, v in rep_units.items()}

    abs_amounts_start_end = np.ones(len(start_end_smiles))
    abs_amounts_ru = {k: 1.0 for k in rep_units.keys()}
    for rep_w, chie in rep_units.items():
        rep_m = total_weight_minus * chie
        rep_n = rep_m / molar_wt_from_smiles(rep_w)
        abs_amounts_ru[rep_w] = rep_n

    total_amounts = sum(abs_amounts_ru.values()) + sum(abs_amounts_start_end)

    return {
        k: v / total_amounts for k, v in abs_amounts_ru.items()
    }, abs_amounts_start_end / total_amounts


def repeating_unit_combinations(
    frag_smiles: List[str], start: Optional[str] = None, end: Optional[str] = None
) -> List[str]:
    """create all possible combinations of repeating units

    Args:
        frag_smiles (List[str]): list of smiles strings for each repeating unit
        start (Optional[str], optional): start of the polymer. Defaults to None.
        end (Optional[str], optional): end of the polymer. Defaults to None.

    Returns:
        List[str]: list of all possible combinations of repeating units as smiles strings

    Example:
        >>> repeating_unnit_combinations(["A", "B"], start="I", end="E")
        ["IAAE", "IBBE", "IABE", "IBAE"]
    """
    frag_smiles_index = np.arange(len(frag_smiles))
    recomb = np.array(
        np.meshgrid(*([frag_smiles_index] * len(frag_smiles_index)))
    ).T.reshape(-1, len(frag_smiles))
    # remove combinations with no -1

    smiles = []
    for r in recomb:
        ruse = r[r >= 0]
        if len(ruse) == 0:
            continue
        smiles.append([frag_smiles[i] for i in r if i != -1])

    recombinations_smiles = ["".join(s) for s in smiles]
    if start:
        recombinations_smiles = [start + s for s in recombinations_smiles]
    if end:
        recombinations_smiles = [s + end for s in recombinations_smiles]
    return recombinations_smiles


def polymol_fom_smiles(smiles: str) -> Mol:
    RDLogger.DisableLog("rdApp.warning")
    mol = MolFromSmiles(smiles)
    RDLogger.EnableLog("rdApp.warning")
    return mol


def test_polymer_smiles(smiles: str) -> bool:
    """Test if a smiles string is a valid monomer representation.
    For this the first and last atom of the molecule must have a missing bond (radical) like [CH2][CH2](C(=O)(OC))

    Args:
        smiles (str): smiles string of the monomer

    Returns:
        bool: True if the smiles string is a valid monomer representation

    Example:
        >>> test_polymer_smiles("CCC(=O)OC") # has no open ends
        False
        >>> test_polymer_smiles("[CH2][CH]C(=O)OC") # has both open ends, but not at the terminal atoms
        False
        >>> test_polymer_smiles("[CH2][CH](C(=O)(OC))") # has both open ends at the terminal atoms
        True
    """

    try:
        basemol = polymol_fom_smiles(smiles)
        # minimum two radicals
        ori_n_radicals = Descriptors.NumRadicalElectrons(basemol)
        if not ori_n_radicals >= 2:
            return False
        if not basemol:
            return False
        if not polymol_fom_smiles(smiles * 3):
            return False

        hmol = polymol_fom_smiles("[H]" + smiles + "[H]")
        if not hmol:
            return False
        if not Descriptors.NumRadicalElectrons(hmol) == ori_n_radicals - 2:
            return False
        return True
    except Exception:
        return False


def test_startgroup(smiles: str) -> bool:
    """Test if a smiles string is a valid start group.
    For this the last atom in the molecule chain resulting from the smiles must
    have a missing bond (radical) like CC[CH2]

    Args:
        smiles (str): smiles string of the start group

    Returns:
        bool: True if the smiles string is a valid start group

    Example:
        >>> test_startgroup("CCC") # has no open ends
        False
        >>> test_startgroup("C[CH]C") # has a radical bot not at the terminal atom
        False
        >>> test_startgroup("C[CH](C)") # has a radical at the terminal atom
        True
    """
    try:
        basemol = polymol_fom_smiles(smiles)
        # minimum two radicals
        ori_n_radicals = Descriptors.NumRadicalElectrons(basemol)
        if not ori_n_radicals >= 1:
            return False
        if not basemol:
            return False

        hmol = polymol_fom_smiles(smiles + "[H]")
        if not hmol:
            return False
        if not Descriptors.NumRadicalElectrons(hmol) == ori_n_radicals - 1:
            return False
        return True
    except Exception:
        return False


def test_endgroup(smiles: str) -> bool:
    """Test if a smiles string is a valid end group.
    For this the first atom in the molecule chain resulting from the smiles must
    have a missing bond (radical) like [CH2]CC

    Args:
        smiles (str): smiles string of the end group

    Returns:
        bool: True if the smiles string is a valid end group

    Example:
        >>> test_endgroup("CCC") # has no open ends
        False
        >>> test_endgroup("C[CH]C") # has a radical bot not at the terminal atom
        False
        >>> test_endgroup("[CH](C)C") # has a radical at the terminal atom
        True
    """
    try:
        basemol = polymol_fom_smiles(smiles)
        # minimum two radicals
        ori_n_radicals = Descriptors.NumRadicalElectrons(basemol)
        if not ori_n_radicals >= 1:
            return False
        if not basemol:
            return False

        hmol = polymol_fom_smiles("[H]" + smiles)
        if not hmol:
            return False
        if not Descriptors.NumRadicalElectrons(hmol) == ori_n_radicals - 1:
            return False
        return True
    except Exception:
        return False


def test_categorical(data: Iterable):
    """Test if the data is categorical or numerical

    Args:
        data (Iterable): data to test

    Raises:
        ValueError: if the data is neither categorical nor numerical
        ValueError: if the data is empty
        ValueError: if the data is not 1-dimensional

    Returns:
        bool: True if the data is categorical, False if the data is numerical

    """

    # make ndarray from data
    data = np.asarray(data)

    if data.ndim > 1:
        raise ValueError("Data must be 1-dimensional")
    if data.size == 0:
        raise ValueError("Data must not be empty")

    # check if data is categorical

    if data.dtype.kind in "SUO":
        return True

    # check if data is numerical
    if data.dtype.kind in "biufc":
        return False

    raise ValueError("Data type not supported")



from typing import Optional, List, Tuple
import numpy as np
from .fingerprints import (
    create_RDKFingerprint,
    merge_bit_fingerprints,
    weight_sum_fingerprints,
    reduce_fp_set,
    apply_reduction_fp_set,
    create_AtomicPairFingerprint,
)
from ._types import FingerprintFunction, PfpData
from .utils import repeating_unit_combinations, calc_polymer_shares


def create_pfp(
    repeating_units: dict[str, float],
    mol_weight: float,
    start: Optional[str] = None,
    end: Optional[str] = None,
    intersection_fp_size: Optional[int] = 2048,
    enhanced_sum_fp_size: Optional[int] = 2048,
    enhanced_fp_functions: Optional[List[FingerprintFunction]] = None,
) -> np.ndarray:
    """Creates a Polyfingerprint

    Args:
        repeating_units (Dict[str: float]): Dictionary containing the
            SMILES representation of each repeating unit and the
            corresponding relative amount.
        mol_weight (float): Molecular weight of the polymer.
        start (Optional[str], optional): start of the polymer. Defaults to hydrogen.
        end (Optional[str], optional): end of the polymer. Defaults to hydrogen.
        intersection_fp_size (Optional[int], optional): Size of the fingerprint
            used to create the intersection fingerprint. Defaults to 2048.
        enhanced_sum_fp_size (Optional[int], optional): Size of the fingerprint
            used to create the enhanced sum fingerprint. Defaults to 2048.
        enhanced_fp_functions (Optional[Callable[[List[str], int], List[np.ndarray]]], optional):
            List of functions used to create the enhanced fingerprints.
            Defaults to [create_RDKFingerprint, create_AtomicPairFingerprint].

    Returns:
        np.ndarray[[-1], float]:
            Polyfingerprint with length intersection_fp_size + (2 * enhanced_sum_fp_size)*len(enhanced_fp_functions)

    """

    if not start:
        start = "[H]"
    if not end:
        end = "[H]"

    if not enhanced_fp_functions:
        enhanced_fp_functions: List[FingerprintFunction] = [
            create_RDKFingerprint,
            create_AtomicPairFingerprint,
        ]

    # create intersection fingerprint
    # create all possible combinations of repeating units
    fragment_recomb = repeating_unit_combinations(
        list(repeating_units.keys()), start=start, end=end
    )

    intersection_fp = merge_bit_fingerprints(
        create_RDKFingerprint(fragment_recomb, fp_size=intersection_fp_size)
    )

    ru_fractions, start_end_share = calc_polymer_shares(
        repeating_units, [start, end], mol_weight
    )

    enhanced_sum_fp = np.concatenate(
        [
            np.concatenate(
                [
                    weight_sum_fingerprints(
                        efpf(
                            list(repeating_units.keys()),
                            fp_size=enhanced_sum_fp_size,
                        ),
                        weights=list(ru_fractions.values()),
                    ),
                    weight_sum_fingerprints(
                        efpf(
                            [
                                start,
                                end,
                            ],
                            fp_size=enhanced_sum_fp_size,
                        ),
                        weights=start_end_share,
                    ),
                ]
            )
            for efpf in enhanced_fp_functions
        ]
    )

    return np.concatenate([intersection_fp, enhanced_sum_fp])


def reduce_pfp_in_dataset(
    pfp_data: List[PfpData], check_correlation: bool = True
) -> Tuple[List[PfpData], dict]:
    for d in pfp_data:
        if "pfp" not in d or d["pfp"] is None:
            raise ValueError("Missing Polyfingerprint in dataset.")

    fingerprints: List[np.ndarray] = [d["pfp"] for d in pfp_data]
    reduced_fps, mask, reference_fp = reduce_fp_set(
        fingerprints, check_correlation=check_correlation
    )
    for d, rfp in zip(pfp_data, reduced_fps):
        d["pfp"] = rfp

    return pfp_data, {
        "mask": mask,
        "reference_fp": reference_fp,
    }


def apply_reduction_to_pfp_in_dataset(
    pfp_data: List[PfpData],
    mask: np.ndarray[[-1], bool],
    reference_fp: np.ndarray[[-1], float],
) -> List[PfpData]:
    for d in pfp_data:
        if "pfp" not in d or d["pfp"] is None:
            raise ValueError("Missing Polyfingerprint in dataset.")

    fingerprints: List[np.ndarray] = [d["pfp"] for d in pfp_data]
    reduced_fps = apply_reduction_fp_set(fingerprints, mask, reference_fp)
    for d, rfp in zip(pfp_data, reduced_fps):
        d["pfp"] = rfp

    return pfp_data



from __future__ import annotations
from typing import List, Tuple, Optional, Dict, Union
import pandas as pd
import numpy as np
from tqdm import tqdm

from ._types import PfpData, FingerprintFunction
from .core import create_pfp
from .utils import test_polymer_smiles, test_endgroup, test_startgroup
from .logger import PFPLOGGER


def data_encoder_picker(
    data: pd.Series,
):
    """Picks the right encoder for the given data."""
    strtype = str(data.dtype)
    if strtype == "object":
        # one hot encoding
        uniques = data.unique().tolist()
        length = len(uniques)

        def _oh_ecoder(x):
            arr = np.zeros(length + 1)
            try:
                index = uniques.index(x)
                arr[index] = 1
            except ValueError:
                arr[-1] = 1
            return arr

        return _oh_ecoder

    elif (
        strtype.startswith("float")
        or strtype.startswith("int")
        or strtype.startswith("bool")
    ):
        return lambda x: np.array([x], dtype=np.float32)

    raise ValueError(f"Data type {strtype} not supported.")


def df_loader(
    df: pd.DataFrame,
    repeating_unit_columns: List[Tuple[str, str]],
    mw_column: str,
    start_group_column: Optional[str] = None,
    end_group_column: Optional[str] = None,
    y: Optional[Union[str, List[str]]] = None,
    intersection_fp_size: int | None = 2048,
    enhanced_sum_fp_size: int | None = 2048,
    enhanced_fp_functions: List[FingerprintFunction] | None = None,
    additional_columns: Optional[List[str]] = None,
):
    """Loads the data to create a Polyfingerprint from a pandas dataframe.

    Args:
        df (str): a pandas dataframe.
        repeating_unit_columns (Tuple(Tuple[str,str])): Tuple of tuples
            containing the column names of the SMILES representation for
            each repeating unit and the corresponding relative amount.
        mw_column (str): Name of the column containing the molecular weight.
        y (Optional[str]): Name of the column containing the target values.
    """
    if additional_columns is None:
        additional_columns = []

    # check df:
    colstofind = [mw_column] + additional_columns
    for smiles, amount in repeating_unit_columns:
        colstofind.append(smiles)
        colstofind.append(amount)

    if y:
        if isinstance(y, str):
            y = [y]
        colstofind.extend(y)
    else:
        y = []
    for _y in y:
        if _y in additional_columns:
            additional_columns.remove(_y)

    if start_group_column:
        colstofind.append(start_group_column)
    if end_group_column:
        colstofind.append(end_group_column)

    # check if all columns are in df

    missing_cols = list(set(colstofind) - set(df.columns))
    if len(missing_cols) > 0:
        raise ValueError(
            f"Column(s) '{', '.join(missing_cols)}' not found in csv file."
        )

    # print all other columns
    for col in df.columns:
        if col not in colstofind:
            print(f"Warning: Column '{col}' not used.")

    # additional_data_types
    additional_data_encoder = {
        col: data_encoder_picker(df[col]) for col in additional_columns
    }

    alldata: List[PfpData] = []
    additional_data_indices = None
    for _, rowdata in tqdm(df.iterrows(), total=len(df), desc="Loading data"):
        repeatingunits: Dict[str, float] = {}
        for smiles, amount in repeating_unit_columns:
            # skip if no smiles or amount are None or NaN
            if (
                not rowdata[smiles]
                or not rowdata[amount]
                or np.isnan(rowdata[amount])
                or np.isnan(rowdata[mw_column])
            ):
                continue
            if rowdata[smiles] in repeatingunits:
                repeatingunits[rowdata[smiles]] += rowdata[amount]
            else:
                repeatingunits[rowdata[smiles]] = rowdata[amount]

        # skip if no repeating units were found
        if not repeatingunits:
            continue

        # check repeating units
        invalid_ru = False
        for ru in repeatingunits.keys():
            if not test_polymer_smiles(ru):
                PFPLOGGER.warning(f"Invalid SMILES string for repeating unit: {ru}")
                invalid_ru = True
        if invalid_ru:
            continue

        # normalize amounts
        total_amount = sum([ru for ru in repeatingunits.values()])
        for ru in repeatingunits.keys():
            repeatingunits[ru] = repeatingunits[ru] / total_amount

        dy = None
        if y:
            dy = rowdata[y].values.astype(np.float32)
            if any(np.isnan(dy)):
                dy = None

        start_group = None
        if start_group_column:
            start_group = rowdata[start_group_column]

        if start_group and not test_startgroup(start_group):
            PFPLOGGER.warning(f"Invalid SMILES string for start group: {start_group}")
            continue

        end_group = None
        if end_group_column:
            end_group = rowdata[end_group_column]

        if end_group and not test_endgroup(end_group):
            PFPLOGGER.warning(f"Invalid SMILES string for end group: {end_group}")
            continue

        _additional_data = [
            additional_data_encoder[col](rowdata[col])
            for col in additional_data_encoder
        ]
        if additional_data_indices is None:
            additional_data_indices = np.array(
                [0] + [len(x) for x in _additional_data]
            ).cumsum()[:-1]
        pfpdat = PfpData(
            repeating_units=repeatingunits,
            y=dy,
            mw=rowdata[mw_column],
            startgroup=start_group,
            endgroup=end_group,
            additional_data_keys=additional_columns,
            additional_data_indices=additional_data_indices,
            additional_data=np.concatenate(_additional_data),
        )
        alldata.append(pfpdat)

    for d in tqdm(alldata, total=len(alldata), desc="Creating Polyfingerprints"):
        d["pfp"] = create_pfp(
            repeating_units=d["repeating_units"],
            mol_weight=d["mw"],
            start=d["startgroup"],
            end=d["endgroup"],
            intersection_fp_size=intersection_fp_size,
            enhanced_sum_fp_size=enhanced_sum_fp_size,
            enhanced_fp_functions=enhanced_fp_functions,
        )
    return alldata


def csv_loader(
    csv: str,
    repeating_unit_columns: List[Tuple[str, str]],
    mw_column: str,
    start_group_column: Optional[str] = None,
    end_group_column: Optional[str] = None,
    y: Optional[Union[str, List[str]]] = None,
    intersection_fp_size: int | None = 2048,
    enhanced_sum_fp_size: int | None = 2048,
    enhanced_fp_functions: List[FingerprintFunction] | None = None,
    additional_columns: Optional[List[str]] = None,
    **kwargs,
):
    """Loads the data to create a Polyfingerprint from a csv file.

    Args:
        csv (str): Path to the csv file.
        repeating_unit_columns (List(Tuple[str,str])): List of tuples
            containing the column names of the SMILES representation for
            each repeating unit and the corresponding relativ amount.
        mw_column (str): Name of the column containing the molecular weight.
        y (Optional[str]): Name of the column containing the target values.

        kwargs: Keyword arguments passed to pandas.read_csv to load
            the csv file, for more information see:
            https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html

    """
    df = pd.read_csv(csv, **kwargs)
    return df_loader(
        df=df,
        repeating_unit_columns=repeating_unit_columns,
        mw_column=mw_column,
        start_group_column=start_group_column,
        end_group_column=end_group_column,
        y=y,
        intersection_fp_size=intersection_fp_size,
        enhanced_sum_fp_size=enhanced_sum_fp_size,
        enhanced_fp_functions=enhanced_fp_functions,
        additional_columns=additional_columns,
    )


def to_input_output_data(fpdata: List[PfpData]) -> Tuple[np.ndarray, np.ndarray]:
    input_data = np.array(
        [np.concatenate([d["pfp"], d["additional_data"]]) for d in fpdata]
    )
    output_data = np.array([d["y"] for d in fpdata])

    return input_data, output_data



import unittest
import numpy as np
from unittest.mock import patch


class TestFingerprintFunctions(unittest.TestCase):
    def test_create_RDKFingerprint(self):
        from polymerfingerprint.fingerprints import create_RDKFingerprint

        smiles = ["CC", "CCC"]
        result = create_RDKFingerprint(smiles)
        self.assertIsInstance(result[0], np.ndarray)
        self.assertEqual(result[0].size, 2048)

        result_with_complement = create_RDKFingerprint(
            ["[CH2][CH](C)"], complement=True
        )
        self.assertTrue(
            np.equal(
                result_with_complement[0], create_RDKFingerprint(["CC(C)"])[0]
            ).all()
        )

    def test_merge_bit_fingerprints(self):
        from polymerfingerprint.fingerprints import merge_bit_fingerprints

        fps = [np.array([True, False, True]), np.array([False, True, True])]
        result = merge_bit_fingerprints(fps)
        self.assertEqual(result.tolist(), [True, True, True])

        with self.assertRaises(ValueError):
            merge_bit_fingerprints(
                [np.array([True, False, True]), np.array([False, True])]
            )

    def test_weight_sum_fingerprints(self):
        from polymerfingerprint.fingerprints import weight_sum_fingerprints

        fps = [np.array([1.0, 2.0, 3.0]), np.array([4.0, 5.0, 6.0])]
        weights = [0.5, 0.5]
        result = weight_sum_fingerprints(fps, weights)
        np.testing.assert_array_equal(result, np.array([2.5, 3.5, 4.5]))

        with self.assertRaises(ValueError):
            weight_sum_fingerprints(fps, [0.5])

        with self.assertRaises(ValueError):
            weight_sum_fingerprints(
                [np.array([1.0, 2.0]), np.array([4.0, 5.0, 6.0])], weights
            )


class TestReduceFPSet(unittest.TestCase):
    @patch("polymerfingerprint.PFPLOGGER.info")  # Mock the "info" method of the logger
    def test_reduce_fp_set(self, mock_logger):
        from polymerfingerprint.fingerprints import reduce_fp_set

        # Sample fingerprints for testing
        fp1 = np.array([0.2, 0.5, 0.2, 0.0])
        fp2 = np.array([0.4, 0.6, 0.4, 0.0])
        fp3 = np.array([0.2, 0.7, 0.2, 0.0])

        reduced_fps, mask, reference_fp = reduce_fp_set([fp1, fp2, fp3])

        # Check that the reduced fingerprints have the expected values
        np.testing.assert_array_equal(
            reduced_fps,
            [np.array([0.2, 0.5]), np.array([0.4, 0.6]), np.array([0.2, 0.7])],
        )

        # Check that the mask is correct
        np.testing.assert_array_equal(mask, np.array([False, False, True, True]))

        # Check that the reference fingerprint is correct
        np.testing.assert_array_equal(reference_fp, fp1)

        # Check the logging output
        mock_logger.assert_called_once_with("reduced size by 50%")


class TestReduceAnotherFPSet(unittest.TestCase):
    # catch the logging output
    @patch("polymerfingerprint.PFPLOGGER.info")  # Mock the "info" method of the logger
    def test_reduce_another_fp_set(self, mock_logger):
        from polymerfingerprint.fingerprints import apply_reduction_fp_set

        # Sample fingerprints, mask, and reference_fp for testing
        fp4 = np.array([0.2, 0.6, 0.2])
        fp5 = np.array([0.2, 0.7, 0.1])
        mask = np.array([True, False, True])
        reference_fp = np.array([0.2, 0.5, 0.1])
        reduced_fps = apply_reduction_fp_set(
            fingerprints=[fp4, fp5], mask=mask, reference_fp=reference_fp
        )

        # Check that the reduced fingerprints have the expected values
        np.testing.assert_array_equal(reduced_fps, [np.array([0.6]), np.array([0.7])])

        # Check the logging output
        mock_logger.assert_called_once_with(
            "mean reduction loss is %.0f%% with the highest loss per fingerprint beeing %.0f%%",
            25.0,
            50.0,
        )


if __name__ == "__main__":
    unittest.main()



from . import loader
from ._types import PfpData
from .core import create_pfp, reduce_pfp_in_dataset, apply_reduction_to_pfp_in_dataset
from .logger import PFPLOGGER
from .fingerprints import reduce_fp_set, apply_reduction_fp_set
from .utils import test_polymer_smiles, test_endgroup, test_startgroup
from . import datareader

__all__ = [
    "loader",
    "PfpData",
    "create_pfp",
    "PFPLOGGER",
    "reduce_pfp_in_dataset",
    "reduce_fp_set",
    "apply_reduction_to_pfp_in_dataset",
    "apply_reduction_fp_set",
    "test_polymer_smiles",
    "test_endgroup",
    "test_startgroup",
    "datareader",
]
